/*
 * HFT font engine for Microwindows
 * Copyright (c) 2011 Hameg Instruments GmbH
 *
 * Supports dynamically loading .hft Hameg Font Fileformat
 *
 * Written by Andr Heinze
 * Bugfixed by Greg Haerr
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include "swap.h"
#include "device.h"
#include "devfont.h"
#include "../drivers/genfont.h"

extern MWPIXELVAL gr_background;
extern MWBOOL gr_usebg;


/* Handling routines for HFT fonts, use MWCOREFONT structure */
static void hft_unloadfont(PMWFONT font);

/* these procs used when font ASCII indexed*/
static MWFONTPROCS pcf_fontprocs =
{
	MWTF_ASCII,
	gen_getfontinfo,
	gen_gettextsize,
	gen_gettextbits,
	hft_unloadfont,
	corefont_drawtext,
	NULL, /* setfontsize */
	NULL, /* setfontrotation */
	NULL, /* setfontattr */
};

/* these procs used when font requires UC16 index*/
static MWFONTPROCS pcf_fontprocs16 =
{
	MWTF_UC16, /* routines expect unicode 16 */
	gen_getfontinfo,
	gen16_gettextsize,
	gen_gettextbits,
	hft_unloadfont,
	gen16_drawtext,
	NULL, /* setfontsize */
	NULL, /* setfontrotation */
	NULL, /* setfontattr */
	NULL, /* duplicate not yet implemented */
};

#define uint8	unsigned char
#define uint16	unsigned short
#define uint32	unsigned int //->Andre auf +long ndern!
#define uint64	unsigned long long
#define int8	signed char
#define int16	signed short
#define int32	signed int //->Andre auf +long ndern!
#define int64	signed long long


typedef struct
{
	uint8	blkOfs[4];				//Offset des Blocks
	uint8	blkSize[4];				//Grsse des Blocks
} hft_fontBlock_type;

typedef struct
{
	//Fontbeschreibung
	uint8	maxwidth[4];			/* max width in pixels*/
	uint8	height[4];				/* height in pixels*/
	uint8	ascent[4];				/* ascent (baseline) height*/
	uint8	firstchar[4];			/* first character in bitmap*/
	uint8	size[4];				/* font size in glyphs*/
	uint8	defaultchar[4];			/* default char (not glyph index)*/
	uint8	bits_size[4]; 			/* # words of MWIMAGEBITS bits*/
} hft_fontDescr_type;

typedef struct
{
	//alle Werte sind in Big Endian abgelegt -> xx[0] -> MSB, xx[3] = LSB (fr int)
	uint8				hftHeadSize[4];			//Gre des Headers
	uint8				hftVersion;				//Versionsnummer
	uint8				hftEndianess;			//Targetendianess des Fonts
	uint8				hftDataSize[4];			//Gre der Daten (Filegre - Headergre)

	hft_fontDescr_type	fontDescr;
	hft_fontBlock_type	fontName;
	hft_fontBlock_type	fontBits;
	hft_fontBlock_type	fontOffsets;
	hft_fontBlock_type	fontWidths;
//	hft_fontBlock_type	fontComprBits;
//	hft_fontBlock_type	fontComprOffsets;
//	hft_fontBlock_type	fontComprWidths;
} hft_header_type;

uint32 get_uint32(uint8 *ptr)
{
	return ((ptr[0]<<24) | (ptr[1]<<16) | (ptr[2]<<8) | ptr[3]);
}

void hft_correctEndianess(unsigned char *src, unsigned int byteCount, unsigned int elemSize)
{
	switch (elemSize)
	{
		case 1:
			break;
		case 2:
			{
				unsigned int i;

				for (i=0; i<byteCount; i+=2)
				{
					unsigned short tmp;

					tmp		= *(unsigned short*)(&src[i]);
#ifdef MW_CPU_BIG_ENDIAN
					//Wert in tmp sollte Big Endian sein, ist aber in Little Endian im Speicher
					src[i + 0]= (tmp >> 0) & 0xFF;	//niedrigen 8 Bit zuerst (sind ja eigentlich die oberen 8 Bit)
					src[i + 1]= (tmp >> 8) & 0xFF;	//dann die hheren  (sind ja eigentlich die unteren 8 Bit)
#else
					//Wert in tmp sollte Little Endian sein, ist aber in Big Endian im Speicher
					src[i + 0]= (tmp >> 8) & 0xFF;	//hheren 8 Bit zuerst (sind ja eigentlich die unteren 8 Bit)
					src[i + 1]= (tmp >> 0) & 0xFF;	//dann die niedrigen (sind ja eigentlich die oberen 8 Bit)
#endif
				}
			}
			break;
		case 4:
			{
				unsigned int i;

				for (i=0; i<byteCount; i+=4)
				{
					unsigned int tmp;

					tmp		= *(unsigned int*)(&src[i]);
#ifdef MW_CPU_BIG_ENDIAN
					//Wert in tmp sollte Big Endian sein, ist aber in Little Endian im Speicher
					src[i + 0]= (tmp >>  0) & 0xFF;
					src[i + 1]= (tmp >>  8) & 0xFF;
					src[i + 2]= (tmp >> 16) & 0xFF;
					src[i + 3]= (tmp >> 24) & 0xFF;
#else
					//Wert in tmp sollte Little Endian sein, ist aber in Big Endian im Speicher
					src[i + 0]= (tmp >> 24) & 0xFF;
					src[i + 1]= (tmp >> 16) & 0xFF;
					src[i + 2]= (tmp >>  8) & 0xFF;
					src[i + 3]= (tmp >>  0) & 0xFF;
#endif
				}
			}
			break;
		default:
			//not done yet
			break;
	}
}

PMWCOREFONT hft_createfont(const char *name, MWCOORD height, int attr)
{
	GdFile		*file = 0;
	MWCOREFONT	*pf = 0;
	int err = 0;
	int uc16;
	int result;
	char fname[256];
	char *p;

	hft_header_type		header;
	uint8	*fontData;
	uint32	headerSize,	dataSize;
	uint32	bitsOfs, 	bitsSize;
	uint32	ofsOfs,		ofsSize;
	uint32	widthOfs,	widthSize;
	uint32	nameOfs;
	int		correctEndianess;

	/* check for pathname prefix*/
	if (strchr(name, '/') != NULL)
		strcpy(fname, name);
	else {
		strcpy(fname, HFT_FONT_DIR "/");
		strcat(fname, name);
	}

	/* check for extension*/
	if ((p = strrchr(fname, '.')) == NULL || strcmp(p, ".hft") != 0) {
		strcat(fname, ".hft");
	}

	//rupfen wir das file mal auf
	file = GdFOpen(fname, "rb");
	if (!file)
	{
		/* check for pathname prefix*/
		if (strchr(name, '/') == NULL) {
			strcpy(fname, HFT_FONT_DIR_2 "/");
			strcat(fname, name);

			/* check for extension*/
			if ((p = strrchr(fname, '.')) == NULL || strcmp(p, ".hft") != 0) {
				strcat(fname, ".hft");
			}

			file = GdFOpen(fname, "rb");
		}
	}

	//wenn nicht gefunzt hat, wech hier...
	if (!file)
		return NULL;

	//zuerts mal die Grsse des Headers einlesen
	result = GdFRead(file, header.hftHeadSize, 4);
	if (result != 4)
	{
		err = -1;
		goto leave_func;
	}
	headerSize = get_uint32(header.hftHeadSize);

	//Headersize wird aufgerundet
	if (headerSize < sizeof(hft_header_type) || headerSize > (sizeof(hft_header_type) + 4))
	{
		err = -1;
		goto leave_func;
	}

	//dann den rest des Headers (ohne die Gre)
	result = GdFRead(file, header.hftHeadSize + 4, sizeof(hft_header_type) - 4);
	if (result < 0 || result != (sizeof(hft_header_type) - 4))
	{
		err = -1;
		goto leave_func;
	}

	//aufgerundete Bytes wegwerfen
	headerSize -= sizeof(hft_header_type);
	while (headerSize > 0)
	{
		if (headerSize < sizeof(fname))
			dataSize = headerSize;
		else
			dataSize = sizeof(fname);

		result = GdFRead(file, fname, dataSize);
		if (result < 0 || result != dataSize)
		{
			err = -1;
			goto leave_func;
		}
		headerSize -= dataSize;
	}

	//sonst die frs nanoX notwendigen Structe mallocen
	if (!(pf = (MWCOREFONT *) GdMalloc(sizeof(MWCOREFONT))))
	{
		err = -1;
		goto leave_func;
	}

	if (!(pf->cfont = (PMWCFONT) GdCalloc(sizeof(MWCFONT), 1)))
	{
		err = -1;
		goto leave_func;
	}

	//Datengre ermitteln und den Rest einlesen
	dataSize = get_uint32(header.hftDataSize);
	if (!(fontData = (uint8*)GdMalloc(dataSize)))
	{
		err = -1;
		goto leave_func;
	}
	result = GdFRead(file, fontData, dataSize);
	if (result < 0 || result != dataSize)
	{
		err = -1;
		goto leave_func;
	}

	bitsOfs		= get_uint32(header.fontBits.blkOfs);
	ofsOfs		= get_uint32(header.fontOffsets.blkOfs);
	widthOfs	= get_uint32(header.fontWidths.blkOfs);
	nameOfs		= get_uint32(header.fontName.blkOfs);

	bitsSize	= get_uint32(header.fontBits.blkSize);
	ofsSize		= get_uint32(header.fontOffsets.blkSize);
	widthSize	= get_uint32(header.fontWidths.blkSize);

	pf->cfont->name		= (char*)(fontData + nameOfs);
	pf->cfont->maxwidth	= get_uint32(header.fontDescr.maxwidth);
	pf->cfont->height	= get_uint32(header.fontDescr.height);
	pf->cfont->ascent	= get_uint32(header.fontDescr.ascent);
	pf->cfont->firstchar= get_uint32(header.fontDescr.firstchar);
	pf->cfont->size		= get_uint32(header.fontDescr.size);

	pf->cfont->bits		= (MWIMAGEBITS *)(fontData + bitsOfs);
	if (ofsSize)
		pf->cfont->offset	= (unsigned long *)(fontData + ofsOfs);
	else
		pf->cfont->offset	= NULL;
	if (widthSize)
		pf->cfont->width	= (uint8*)(fontData + widthOfs);
	else
		pf->cfont->width	= NULL;

	correctEndianess = FALSE;
#ifdef MW_CPU_BIG_ENDIAN
	if (!header.hftEndianess)
		correctEndianess = TRUE;
#else
	if (header.hftEndianess)
		correctEndianess = TRUE;
#endif
	if (correctEndianess)
	{
		if (pf->cfont->bits)
			hft_correctEndianess((uint8*)pf->cfont->bits , bitsSize, sizeof(MWIMAGEBITS));
		if (pf->cfont->offset)
			hft_correctEndianess((uint8*)pf->cfont->offset , ofsSize, sizeof(unsigned long));
	}

	pf->cfont->defaultchar	= get_uint32(header.fontDescr.defaultchar);
	pf->cfont->bits_size	= get_uint32(header.fontDescr.bits_size);

	uc16 = pf->cfont->firstchar > 255 || (pf->cfont->firstchar + pf->cfont->size) > 256;
	pf->fontprocs = uc16 ? &pcf_fontprocs16 : &pcf_fontprocs;
	pf->fontsize = pf->fontrotation = pf->fontattr = 0;
	pf->name = "HFT";

leave_func:
	if (file)
		GdFClose(file);

	if (err == 0 && pf)
		return pf;

	hft_unloadfont((PMWFONT) pf);
	return 0;
}

void hft_unloadfont(PMWFONT font)
{
	PMWCOREFONT pf = (PMWCOREFONT) font;
	PMWCFONT pfc = pf->cfont;

	if (pfc)
	{
		if (pfc->width)
			GdFree(pf->cfont->width);
		if (pfc->offset)
			GdFree(pf->cfont->offset);
		if (pfc->bits)
			GdFree(pf->cfont->bits);

		GdFree2(pf->cfont, sizeof(MWCFONT));
	}

	GdFree2(font, sizeof(MWCOREFONT));
}
